"""
Migration script to add the new_repo_alert column to the galaxy_user table.
"""

import logging
import sys

from sqlalchemy import (
    Boolean,
    Column,
    MetaData,
    Table,
)

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Create and initialize imported column in job table.
    User_table = Table("galaxy_user", metadata, autoload=True)
    c = Column("new_repo_alert", Boolean, default=False, index=True)
    try:
        # Create
        c.create(User_table, index_name="ix_galaxy_user_new_repo_alert")
        assert c is User_table.c.new_repo_alert
        # Initialize.
        if migrate_engine.name == "mysql" or migrate_engine.name == "sqlite":
            default_false = "0"
        elif migrate_engine.name in ["postgresql", "postgres"]:
            default_false = "false"
        else:
            log.debug("unknown migrate_engine dialect")
        migrate_engine.execute(f"UPDATE galaxy_user SET new_repo_alert={default_false}")
    except Exception:
        log.exception("Adding new_repo_alert column to the galaxy_user table failed.")


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop new_repo_alert column from galaxy_user table.
    User_table = Table("galaxy_user", metadata, autoload=True)
    try:
        User_table.c.new_repo_alert.drop()
    except Exception:
        log.exception("Dropping column new_repo_alert from the galaxy_user table failed.")



"""Migration script to add the numeric_revision column to the repository metadata table."""

import logging
import sys

from sqlalchemy import (
    Column,
    Integer,
    MetaData,
    Table,
)

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)
    c = Column("numeric_revision", Integer, index=True)
    try:
        # Create
        c.create(RepositoryMetadata_table, index_name="ix_numeric_revision")
        assert c is RepositoryMetadata_table.c.numeric_revision
    except Exception:
        log.exception("Adding numeric_revision column to the repository table failed.")
    # Update the numeric_revision column to have the default undefined value.
    cmd = "UPDATE repository_metadata SET numeric_revision = -1"
    migrate_engine.execute(cmd)


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop numeric_revision column from repository table.
    RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)
    try:
        RepositoryMetadata_table.c.numeric_revision.drop()
    except Exception:
        log.exception("Dropping column numeric_revision from the repository table failed.")



"""
Migration script to add the repository_review, component_review and component tables and the Repository Reviewer group and role.
"""

import datetime
import logging
import sys

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    Table,
    TEXT,
)

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import TrimmedString

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()

IUC = "Intergalactic Utilities Commission"
NOW = datetime.datetime.utcnow
REVIEWER = "Repository Reviewer"
ROLE_TYPE = "system"


def nextval(migrate_engine, table, col="id"):
    if migrate_engine.name in ["postgresql", "postgres"]:
        return f"nextval('{table}_{col}_seq')"
    elif migrate_engine.name == "mysql" or migrate_engine.name == "sqlite":
        return "null"
    else:
        raise Exception(f"Unable to convert data for unknown database type: {migrate_engine.name}")


def localtimestamp(migrate_engine):
    if migrate_engine.name in ["postgresql", "postgres"] or migrate_engine.name == "mysql":
        return "LOCALTIMESTAMP"
    elif migrate_engine.name == "sqlite":
        return "current_date || ' ' || current_time"
    else:
        raise Exception(f"Unable to convert data for unknown database type: {migrate_engine.name}")


def boolean_false(migrate_engine):
    if migrate_engine.name in ["postgresql", "postgres"] or migrate_engine.name == "mysql":
        return False
    elif migrate_engine.name == "sqlite":
        return 0
    else:
        raise Exception(f"Unable to convert data for unknown database type: {migrate_engine.name}")


RepositoryReview_table = Table(
    "repository_review",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=NOW),
    Column("update_time", DateTime, default=NOW, onupdate=NOW),
    Column("repository_id", Integer, ForeignKey("repository.id"), index=True),
    Column("changeset_revision", TrimmedString(255), index=True),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True, nullable=False),
    Column("approved", TrimmedString(255)),
    Column("rating", Integer, index=True),
    Column("deleted", Boolean, index=True, default=False),
)

ComponentReview_table = Table(
    "component_review",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=NOW),
    Column("update_time", DateTime, default=NOW, onupdate=NOW),
    Column("repository_review_id", Integer, ForeignKey("repository_review.id"), index=True),
    Column("component_id", Integer, ForeignKey("component.id"), index=True),
    Column("comment", TEXT),
    Column("private", Boolean, default=False),
    Column("approved", TrimmedString(255)),
    Column("rating", Integer),
    Column("deleted", Boolean, index=True, default=False),
)

Component_table = Table(
    "component",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("name", TrimmedString(255)),
    Column("description", TEXT),
)


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Create new review tables.
    try:
        Component_table.create()
    except Exception:
        log.exception("Creating component table failed.")
    try:
        RepositoryReview_table.create()
    except Exception:
        log.exception("Creating repository_review table failed.")
    try:
        ComponentReview_table.create()
    except Exception:
        log.exception("Creating component_review table failed.")
    # Insert default Component values.
    names = ["Data types", "Functional tests", "README", "Tool dependencies", "Tools", "Workflows"]
    descriptions = [
        "Proprietary datatypes defined in a file named datatypes_conf.xml included in the repository",
        "Functional tests defined in each tool config included in the repository along with test data files",
        "An appropriately named file included in the repository that contains installation information or 3rd-party tool dependency licensing information",
        "Tool dependencies defined in a file named tool_dependencies.xml included in the repository for contained tools",
        "Galaxy tools included in the repository",
        "Exported Galaxy workflows included in the repository",
    ]
    for tup in zip(names, descriptions):
        name, description = tup
        cmd = "INSERT INTO component VALUES ("
        cmd += f"{nextval(migrate_engine, 'component')}, "
        cmd += f"'{name}', "
        cmd += f"'{description}' "
        cmd += ");"
        migrate_engine.execute(cmd)
    # Insert a REVIEWER role into the role table.
    cmd = "INSERT INTO role VALUES ("
    cmd += f"{nextval(migrate_engine, 'role')}, "
    cmd += f"{localtimestamp(migrate_engine)}, "
    cmd += f"{localtimestamp(migrate_engine)}, "
    cmd += f"'{REVIEWER}', "
    cmd += "'A user or group member with this role can review repositories.', "
    cmd += f"'{ROLE_TYPE}', "
    cmd += f"{boolean_false(migrate_engine)}"
    cmd += ");"
    migrate_engine.execute(cmd)
    # Get the id of the REVIEWER role.
    cmd = f"SELECT id FROM role WHERE name = '{REVIEWER}' and type = '{ROLE_TYPE}';"
    row = migrate_engine.execute(cmd).fetchone()
    if row:
        role_id = row[0]
    else:
        role_id = None
    # Insert an IUC group into the galaxy_group table.
    cmd = "INSERT INTO galaxy_group VALUES ("
    cmd += f"{nextval(migrate_engine, 'galaxy_group')}, "
    cmd += f"{localtimestamp(migrate_engine)}, "
    cmd += f"{localtimestamp(migrate_engine)}, "
    cmd += f"'{IUC}', "
    cmd += f"{boolean_false(migrate_engine)}"
    cmd += ");"
    migrate_engine.execute(cmd)
    # Get the id of the IUC group.
    cmd = f"SELECT id FROM galaxy_group WHERE name = '{IUC}';"
    row = migrate_engine.execute(cmd).fetchone()
    if row:
        group_id = row[0]
    else:
        group_id = None
    if group_id and role_id:
        # Insert a group_role_association for the IUC group and the REVIEWER role.
        cmd = "INSERT INTO group_role_association VALUES ("
        cmd += f"{nextval(migrate_engine, 'group_role_association')}, "
        cmd += "%d, " % int(group_id)
        cmd += "%d, " % int(role_id)
        cmd += f"{localtimestamp(migrate_engine)}, "
        cmd += f"{localtimestamp(migrate_engine)} "
        cmd += ");"
        migrate_engine.execute(cmd)


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop review tables.
    try:
        ComponentReview_table.drop()
    except Exception:
        log.exception("Dropping component_review table failed.")
    try:
        RepositoryReview_table.drop()
    except Exception:
        log.exception("Dropping repository_review table failed.")
    try:
        Component_table.drop()
    except Exception:
        log.exception("Dropping component table failed.")
    # Get the id of the REVIEWER group.
    cmd = f"SELECT id FROM galaxy_group WHERE name = '{IUC}';"
    row = migrate_engine.execute(cmd).fetchone()
    if row:
        group_id = row[0]
    else:
        group_id = None
    # Get the id of the REVIEWER role.
    cmd = f"SELECT id FROM role WHERE name = '{REVIEWER}' and type = '{ROLE_TYPE}';"
    row = migrate_engine.execute(cmd).fetchone()
    if row:
        role_id = row[0]
    else:
        role_id = None
    # See if we have at least 1 user
    cmd = "SELECT * FROM galaxy_user;"
    users = migrate_engine.execute(cmd).fetchall()
    if role_id:
        if users:
            # Delete all UserRoleAssociations for the REVIEWER role.
            cmd = "DELETE FROM user_role_association WHERE role_id = %d;" % int(role_id)
            migrate_engine.execute(cmd)
        if group_id:
            # Delete all UserGroupAssociations for members of the IUC group.
            cmd = "DELETE FROM user_group_association WHERE group_id = %d;" % int(group_id)
            migrate_engine.execute(cmd)
            # Delete all GroupRoleAssociations for the IUC group and the REVIEWER role.
            cmd = "DELETE FROM group_role_association WHERE group_id = %d and role_id = %d;" % (
                int(group_id),
                int(role_id),
            )
            migrate_engine.execute(cmd)
            # Delete the IUC group from the galaxy_group table.
            cmd = "DELETE FROM galaxy_group WHERE id = %d;" % int(group_id)
            migrate_engine.execute(cmd)
        # Delete the REVIEWER role from the role table.
        cmd = "DELETE FROM role WHERE id = %d;" % int(role_id)
        migrate_engine.execute(cmd)



import logging
import os.path
import sys

from migrate.versioning import (
    repository,
    schema,
)
from sqlalchemy import (
    create_engine,
    MetaData,
    Table,
)
from sqlalchemy.exc import NoSuchTableError

from galaxy.model.database_utils import (
    create_database,
    database_exists,
)

log = logging.getLogger(__name__)

# path relative to galaxy
migrate_repository_directory = os.path.dirname(__file__).replace(os.getcwd() + os.path.sep, "", 1)
migrate_repository = repository.Repository(migrate_repository_directory)


def create_or_verify_database(url, engine_options=None):
    """
    Check that the database is useable, possibly creating it if empty (this is
    the only time we automatically create tables, otherwise we force the
    user to do it using the management script so they can create backups).

    1) Empty database --> initialize with latest version and return
    2) Database older than migration support --> fail and require manual update
    3) Database at state where migrate support introduced --> add version control information but make no changes (might still require manual update)
    4) Database versioned but out of date --> fail with informative message, user must run "sh manage_toolshed_db.sh upgrade"

    """
    engine_options = engine_options or {}
    # Create engine and metadata
    if not database_exists(url):
        message = f"Creating database for URI [{url}]"
        log.info(message)
        create_database(url)

    engine = create_engine(url, **engine_options)
    meta = MetaData(bind=engine)
    # Try to load dataset table
    try:
        Table("galaxy_user", meta, autoload=True)
    except NoSuchTableError:
        # No 'galaxy_user' table means a completely uninitialized database, which
        # is fine, init the database in a versioned state
        log.info("No database, initializing")
        # Database might or might not be versioned
        try:
            # Declare the database to be under a repository's version control
            db_schema = schema.ControlledSchema.create(engine, migrate_repository)
        except Exception:
            # The database is already under version control
            db_schema = schema.ControlledSchema(engine, migrate_repository)
        # Apply all scripts to get to current version
        migrate_to_current_version(engine, db_schema)
        return
    try:
        Table("migrate_version", meta, autoload=True)
    except NoSuchTableError:
        # The database exists but is not yet under migrate version control, so init with version 1
        log.info("Adding version control to existing database")
        try:
            Table("metadata_file", meta, autoload=True)
            schema.ControlledSchema.create(engine, migrate_repository, version=2)
        except NoSuchTableError:
            schema.ControlledSchema.create(engine, migrate_repository, version=1)
    # Verify that the code and the DB are in sync
    db_schema = schema.ControlledSchema(engine, migrate_repository)
    if migrate_repository.versions.latest != db_schema.version:
        exception_msg = "Your database has version '%d' but this code expects version '%d'.  " % (
            db_schema.version,
            migrate_repository.versions.latest,
        )
        exception_msg += "Back up your database and then migrate the schema by running the following from your Galaxy installation directory:"
        exception_msg += "\n\nsh manage_toolshed_db.sh upgrade\n"
        raise Exception(exception_msg)
    else:
        log.info("At database version %d" % db_schema.version)


def migrate_to_current_version(engine, schema):
    # Changes to get to current version
    changeset = schema.changeset(None)
    for ver, change in changeset:
        nextver = ver + changeset.step
        log.info(f"Migrating {ver} -> {nextver}... ")
        old_stdout = sys.stdout

        class FakeStdout:
            def __init__(self):
                self.buffer = []

            def write(self, s):
                self.buffer.append(s)

            def flush(self):
                pass

        sys.stdout = FakeStdout()
        try:
            schema.runchange(ver, change, changeset.step)
        finally:
            for message in "".join(sys.stdout.buffer).split("\n"):
                log.info(message)
            sys.stdout = old_stdout



"""
Migration script to add the tool_versions column to the repository_metadata table.
"""

import datetime
import logging
import sys

from sqlalchemy import (
    Column,
    MetaData,
    Table,
)

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import JSONType

now = datetime.datetime.utcnow
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)
    c = Column("tool_versions", JSONType, nullable=True)
    try:
        # Create
        c.create(RepositoryMetadata_table)
        assert c is RepositoryMetadata_table.c.tool_versions
    except Exception:
        log.exception("Adding tool_versions column to the repository_metadata table failed.")


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop new_repo_alert column from galaxy_user table.
    RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)
    try:
        RepositoryMetadata_table.c.tool_versions.drop()
    except Exception:
        log.exception("Dropping column tool_versions from the repository_metadata table failed.")



"""
Migration script for the password reset table
"""

import datetime
import logging

from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    String,
    Table,
)

now = datetime.datetime.utcnow
log = logging.getLogger(__name__)
metadata = MetaData()

PasswordResetToken_table = Table(
    "password_reset_token",
    metadata,
    Column("token", String(32), primary_key=True, unique=True, index=True),
    Column("expiration_time", DateTime),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
)


def upgrade(migrate_engine):
    metadata.bind = migrate_engine
    print(__doc__)
    metadata.reflect()
    try:
        PasswordResetToken_table.create()
    except Exception:
        log.exception("Creating %s table failed", PasswordResetToken_table.name)


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    try:
        PasswordResetToken_table.drop()
    except Exception:
        log.exception("Dropping %s table failed", PasswordResetToken_table.name)



"""
Migration script to add deleted column to API keys
"""

import datetime
import logging

from sqlalchemy import (
    Boolean,
    Column,
    MetaData,
    Table,
)

now = datetime.datetime.utcnow
log = logging.getLogger(__name__)
metadata = MetaData()

TABLE_NAME = "api_keys"
COLUMN_NAME = "deleted"


def upgrade(migrate_engine):
    metadata.bind = migrate_engine
    print(__doc__)
    metadata.reflect()

    deleted_column = Column(COLUMN_NAME, Boolean)
    __add_column(deleted_column, TABLE_NAME, metadata)


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()

    __drop_column(COLUMN_NAME, TABLE_NAME, metadata)


def __add_column(column, table_name, metadata, **kwds):
    try:
        table = Table(table_name, metadata, autoload=True)
        column.create(table, **kwds)
    except Exception:
        log.exception("Adding column %s failed.", column)


def __drop_column(column_name, table_name, metadata):
    try:
        table = Table(table_name, metadata, autoload=True)
        getattr(table.c, column_name).drop()
    except Exception:
        log.exception("Dropping column %s failed.", column_name)



from galaxy.tool_shed.util.tool_util import (
    copy_sample_file,
    copy_sample_files,
    generate_message_for_invalid_tools,
    new_state,
)

__all__ = (
    "copy_sample_file",
    "copy_sample_files",
    "generate_message_for_invalid_tools",
    "new_state",
)



"""
Drops the tool, tool_category_association, event, tool_event_association, tool_rating_association,
tool_tag_association and tool_annotation_association tables since they are no longer used in the
next-gen tool shed.
"""

import datetime
import logging
import sys

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    Table,
    TEXT,
)
from sqlalchemy.exc import NoSuchTableError

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import TrimmedString

now = datetime.datetime.utcnow
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    # Load existing tables
    metadata.bind = migrate_engine
    metadata.reflect()
    # Load and then drop the tool_category_association table
    try:
        ToolCategoryAssociation_table = Table("tool_category_association", metadata, autoload=True)
    except NoSuchTableError:
        log.debug("Failed loading table tool_category_association")
    try:
        ToolCategoryAssociation_table.drop()
    except Exception:
        log.exception("Dropping tool_category_association table failed.")
    # Load and then drop the tool_event_association table
    try:
        ToolEventAssociation_table = Table("tool_event_association", metadata, autoload=True)
    except NoSuchTableError:
        log.debug("Failed loading table tool_event_association")
    try:
        ToolEventAssociation_table.drop()
    except Exception:
        log.exception("Dropping tool_event_association table failed.")
    # Load and then drop the tool_rating_association table
    try:
        ToolRatingAssociation_table = Table("tool_rating_association", metadata, autoload=True)
    except NoSuchTableError:
        log.debug("Failed loading table tool_rating_association")
    try:
        ToolRatingAssociation_table.drop()
    except Exception:
        log.exception("Dropping tool_rating_association table failed.")
    # Load and then drop the tool_tag_association table
    try:
        ToolTagAssociation_table = Table("tool_tag_association", metadata, autoload=True)
    except NoSuchTableError:
        log.debug("Failed loading table tool_tag_association")
    try:
        ToolTagAssociation_table.drop()
    except Exception:
        log.exception("Dropping tool_tag_association table failed.")
    # Load and then drop the tool_annotation_association table
    try:
        ToolAnnotationAssociation_table = Table("tool_annotation_association", metadata, autoload=True)
    except NoSuchTableError:
        log.debug("Failed loading table tool_annotation_association")
    try:
        ToolAnnotationAssociation_table.drop()
    except Exception:
        log.exception("Dropping tool_annotation_association table failed.")
    # Load and then drop the event table
    try:
        Event_table = Table("event", metadata, autoload=True)
    except NoSuchTableError:
        log.debug("Failed loading table event")
    try:
        Event_table.drop()
    except Exception:
        log.exception("Dropping event table failed.")
    # Load and then drop the tool table
    try:
        Tool_table = Table("tool", metadata, autoload=True)
    except NoSuchTableError:
        log.debug("Failed loading table tool")
    try:
        Tool_table.drop()
    except Exception:
        log.exception("Dropping tool table failed.")


def downgrade(migrate_engine):
    # Load existing tables
    metadata.bind = migrate_engine
    metadata.reflect()
    # We've lost all of our data, so downgrading is useless. However, we'll
    # at least re-create the dropped tables.
    Event_table = Table(
        "event",
        metadata,
        Column("id", Integer, primary_key=True),
        Column("create_time", DateTime, default=now),
        Column("update_time", DateTime, default=now, onupdate=now),
        Column("state", TrimmedString(255), index=True),
        Column("comment", TEXT),
    )

    Tool_table = Table(
        "tool",
        metadata,
        Column("id", Integer, primary_key=True),
        Column("guid", TrimmedString(255), index=True, unique=True),
        Column("tool_id", TrimmedString(255), index=True),
        Column("create_time", DateTime, default=now),
        Column("update_time", DateTime, default=now, onupdate=now),
        Column("newer_version_id", Integer, ForeignKey("tool.id"), nullable=True),
        Column("name", TrimmedString(255), index=True),
        Column("description", TEXT),
        Column("user_description", TEXT),
        Column("version", TrimmedString(255)),
        Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
        Column("external_filename", TEXT),
        Column("deleted", Boolean, index=True, default=False),
        Column("suite", Boolean, default=False, index=True),
    )

    ToolCategoryAssociation_table = Table(
        "tool_category_association",
        metadata,
        Column("id", Integer, primary_key=True),
        Column("tool_id", Integer, ForeignKey("tool.id"), index=True),
        Column("category_id", Integer, ForeignKey("category.id"), index=True),
    )

    ToolEventAssociation_table = Table(
        "tool_event_association",
        metadata,
        Column("id", Integer, primary_key=True),
        Column("tool_id", Integer, ForeignKey("tool.id"), index=True),
        Column("event_id", Integer, ForeignKey("event.id"), index=True),
    )

    ToolRatingAssociation_table = Table(
        "tool_rating_association",
        metadata,
        Column("id", Integer, primary_key=True),
        Column("create_time", DateTime, default=now),
        Column("update_time", DateTime, default=now, onupdate=now),
        Column("tool_id", Integer, ForeignKey("tool.id"), index=True),
        Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
        Column("rating", Integer, index=True),
        Column("comment", TEXT),
    )

    ToolTagAssociation_table = Table(
        "tool_tag_association",
        metadata,
        Column("id", Integer, primary_key=True),
        Column("tool_id", Integer, ForeignKey("tool.id"), index=True),
        Column("tag_id", Integer, ForeignKey("tag.id"), index=True),
        Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
        Column("user_tname", TrimmedString(255), index=True),
        Column("value", TrimmedString(255), index=True),
        Column("user_value", TrimmedString(255), index=True),
    )

    ToolAnnotationAssociation_table = Table(
        "tool_annotation_association",
        metadata,
        Column("id", Integer, primary_key=True),
        Column("tool_id", Integer, ForeignKey("tool.id"), index=True),
        Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
        Column("annotation", TEXT, index=True),
    )

    # Create the event table
    try:
        Event_table.create()
    except Exception:
        log.exception("Creating event table failed.")
    # Create the tool table
    try:
        Tool_table.create()
    except Exception:
        log.exception("Creating tool table failed.")
    # Create the tool_category_association table
    try:
        ToolCategoryAssociation_table.create()
    except Exception:
        log.exception("Creating tool_category_association table failed.")
    # Create the tool_event_association table
    try:
        ToolEventAssociation_table.create()
    except Exception:
        log.exception("Creating tool_event_association table failed.")
    # Create the tool_rating_association table
    try:
        ToolRatingAssociation_table.create()
    except Exception:
        log.exception("Creating tool_rating_association table failed.")
    # Create the tool_tag_association table
    try:
        ToolTagAssociation_table.create()
    except Exception:
        log.exception("Creating tool_tag_association table failed.")
    # Create the tool_annotation_association table
    try:
        ToolAnnotationAssociation_table.create()
    except Exception:
        log.exception("Creating tool_annotation_association table failed.")



"""
API operations for uploaded files in storage.
"""
import logging
import os
import re

from galaxy import exceptions
from galaxy.web.framework.decorators import (
    expose_api_raw_anonymous,
    legacy_expose_api_anonymous,
)
from . import BaseGalaxyAPIController

log = logging.getLogger(__name__)


class UploadsAPIController(BaseGalaxyAPIController):
    READ_CHUNK_SIZE = 2**16

    @expose_api_raw_anonymous
    def hooks(self, trans, **kwds):
        """
        Exposed as POST /api/upload/hooks and /api/upload/resumable_upload
        """
        # Internal endpoint, only purpose is to authenticate user, but may grow additional functionality in the future
        return None

    @legacy_expose_api_anonymous
    def index(self, trans, **kwd):
        raise exceptions.NotImplemented("Listing uploads is not implemented.")

    @legacy_expose_api_anonymous
    def create(self, trans, payload, **kwd):
        """
        POST /api/uploads/
        """
        session_id = payload.get("session_id")
        session_start = payload.get("session_start")
        session_chunk = payload.get("session_chunk")
        if re.match(r"^[\w-]+$", session_id) is None:
            raise exceptions.MessageException("Requires a session id.")
        if session_start is None:
            raise exceptions.MessageException("Requires a session start.")
        if not hasattr(session_chunk, "file"):
            raise exceptions.MessageException("Requires a session chunk.")
        target_file = os.path.join(trans.app.config.new_file_path, session_id)
        target_size = 0
        if os.path.exists(target_file):
            target_size = os.path.getsize(target_file)
        if session_start != target_size:
            raise exceptions.MessageException("Incorrect session start.")
        chunk_size = os.fstat(session_chunk.file.fileno()).st_size
        if chunk_size > trans.app.config.chunk_upload_size:
            raise exceptions.MessageException("Invalid chunk size.")
        with open(target_file, "ab") as f:
            while True:
                read_chunk = session_chunk.file.read(self.READ_CHUNK_SIZE)
                if not read_chunk:
                    break
                f.write(read_chunk)
        session_chunk.file.close()
        return {"message": "Successful."}



"""
Migration script to add the repository_metadata table.
"""

import datetime
import logging
import sys

from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    Table,
)

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import (
    JSONType,
    TrimmedString,
)

now = datetime.datetime.utcnow
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()

RepositoryMetadata_table = Table(
    "repository_metadata",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("repository_id", Integer, ForeignKey("repository.id"), index=True),
    Column("changeset_revision", TrimmedString(255), index=True),
    Column("metadata", JSONType, nullable=True),
)


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Create repository_metadata table.
    try:
        RepositoryMetadata_table.create()
    except Exception:
        log.exception("Creating repository_metadata table failed.")


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop repository_metadata table.
    try:
        RepositoryMetadata_table.drop()
    except Exception:
        log.exception("Dropping repository_metadata table failed.")



"""
Migration script to add the long_description and times_downloaded columns to the repository table.
"""

import logging
import sys

from sqlalchemy import (
    Column,
    Integer,
    MetaData,
    Table,
    TEXT,
)

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Create and initialize imported column in job table.
    Repository_table = Table("repository", metadata, autoload=True)
    c = Column("long_description", TEXT)
    try:
        # Create
        c.create(Repository_table)
        assert c is Repository_table.c.long_description
    except Exception:
        log.exception("Adding long_description column to the repository table failed.")

    c = Column("times_downloaded", Integer)
    try:
        # Create
        c.create(Repository_table)
        assert c is Repository_table.c.times_downloaded
    except Exception:
        log.exception("Adding times_downloaded column to the repository table failed.")

    cmd = "UPDATE repository SET long_description = ''"
    migrate_engine.execute(cmd)
    cmd = "UPDATE repository SET times_downloaded = 0"
    migrate_engine.execute(cmd)


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop email_alerts column from repository table.
    Repository_table = Table("repository", metadata, autoload=True)
    try:
        Repository_table.c.long_description.drop()
    except Exception:
        log.exception("Dropping column long_description from the repository table failed.")
    try:
        Repository_table.c.times_downloaded.drop()
    except Exception:
        log.exception("Dropping column times_downloaded from the repository table failed.")



"""
Migration script to add the tool_test_errors, do_not_test, tools_functionally_correct, and time_last_tested columns to the repository_metadata table.
"""

import logging
import sys

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    MetaData,
    Table,
)

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import JSONType

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Create and initialize tools_functionally_correct, do_not_test, time_last_tested, and tool_test_errors columns in repository_metadata table.
    RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)
    c = Column("tools_functionally_correct", Boolean, default=False, index=True)
    try:
        # Create tools_functionally_correct column
        c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_tfc")
        assert c is RepositoryMetadata_table.c.tools_functionally_correct
        # Initialize.
        if migrate_engine.name == "mysql" or migrate_engine.name == "sqlite":
            default_false = "0"
        elif migrate_engine.name in ["postgresql", "postgres"]:
            default_false = "false"
        migrate_engine.execute(f"UPDATE repository_metadata SET tools_functionally_correct={default_false}")
    except Exception:
        log.exception("Adding tools_functionally_correct column to the repository_metadata table failed.")
    c = Column("do_not_test", Boolean, default=False, index=True)
    try:
        # Create do_not_test column
        c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_dnt")
        assert c is RepositoryMetadata_table.c.do_not_test
        # Initialize.
        if migrate_engine.name == "mysql" or migrate_engine.name == "sqlite":
            default_false = "0"
        elif migrate_engine.name in ["postgresql", "postgres"]:
            default_false = "false"
        migrate_engine.execute(f"UPDATE repository_metadata SET do_not_test={default_false}")
    except Exception:
        log.exception("Adding do_not_test column to the repository_metadata table failed.")
    c = Column("time_last_tested", DateTime, default=None, nullable=True)
    try:
        # Create time_last_tested column
        c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_tlt")
        assert c is RepositoryMetadata_table.c.time_last_tested
    except Exception:
        log.exception("Adding time_last_tested column to the repository_metadata table failed.")
    c = Column("tool_test_errors", JSONType, nullable=True)
    try:
        # Create tool_test_errors column
        c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_tte")
        assert c is RepositoryMetadata_table.c.tool_test_errors
    except Exception:
        log.exception("Adding tool_test_errors column to the repository_metadata table failed.")


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop tool_test_errors, time_last_tested, do_not_test, and tools_functionally_correct columns from repository_metadata table.
    RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)
    try:
        RepositoryMetadata_table.c.tool_test_errors.drop()
    except Exception:
        log.exception("Dropping column tool_test_errors from the repository_metadata table failed.")
    try:
        RepositoryMetadata_table.c.time_last_tested.drop()
    except Exception:
        log.exception("Dropping column time_last_tested from the repository_metadata table failed.")
    try:
        RepositoryMetadata_table.c.do_not_test.drop()
    except Exception:
        log.exception("Dropping column do_not_test from the repository_metadata table failed.")
    try:
        RepositoryMetadata_table.c.tools_functionally_correct.drop()
    except Exception:
        log.exception("Dropping column tools_functionally_correct from the repository_metadata table failed.")



"""Migration script to change repository.type column value from generic to unrestricted."""

import logging
import sys

from sqlalchemy import MetaData

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Update the type column to have the default unrestricted value.
    cmd = "UPDATE repository SET type = 'unrestricted' WHERE type = 'generic'"
    migrate_engine.execute(cmd)


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Update the type column to have the default generic value.
    cmd = "UPDATE repository SET type = 'generic' WHERE type = 'unrestricted'"
    migrate_engine.execute(cmd)



"""
Adds the tool_rating_association table, enabling tools to be rated along with review comments.
"""

import datetime
import logging
import sys

from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    Table,
    TEXT,
)

now = datetime.datetime.utcnow
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()

ToolRatingAssociation_table = Table(
    "tool_rating_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("tool_id", Integer, ForeignKey("tool.id"), index=True),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
    Column("rating", Integer, index=True),
    Column("comment", TEXT),
)


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    # Load existing tables
    metadata.reflect()
    try:
        ToolRatingAssociation_table.create()
    except Exception:
        log.exception("Creating tool_rating_association table failed.")


def downgrade(migrate_engine):
    # Load existing tables
    metadata.bind = migrate_engine
    metadata.reflect()
    try:
        ToolRatingAssociation_table.drop()
    except Exception:
        log.exception("Dropping tool_rating_association table failed.")



"""
Migration script to create initial tables.
"""

import datetime
import logging
import sys

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    MetaData,
    String,
    Table,
    TEXT,
    UniqueConstraint,
)

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import TrimmedString

now = datetime.datetime.utcnow
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()

User_table = Table(
    "galaxy_user",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("email", TrimmedString(255), nullable=False),
    Column("username", String(255), index=True),
    Column("password", TrimmedString(40), nullable=False),
    Column("external", Boolean, default=False),
    Column("deleted", Boolean, index=True, default=False),
    Column("purged", Boolean, index=True, default=False),
)

Group_table = Table(
    "galaxy_group",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("name", String(255), index=True, unique=True),
    Column("deleted", Boolean, index=True, default=False),
)

Role_table = Table(
    "role",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("name", String(255), index=True, unique=True),
    Column("description", TEXT),
    Column("type", String(40), index=True),
    Column("deleted", Boolean, index=True, default=False),
)

UserGroupAssociation_table = Table(
    "user_group_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
    Column("group_id", Integer, ForeignKey("galaxy_group.id"), index=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
)

UserRoleAssociation_table = Table(
    "user_role_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
    Column("role_id", Integer, ForeignKey("role.id"), index=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
)

GroupRoleAssociation_table = Table(
    "group_role_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("group_id", Integer, ForeignKey("galaxy_group.id"), index=True),
    Column("role_id", Integer, ForeignKey("role.id"), index=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
)

GalaxySession_table = Table(
    "galaxy_session",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True, nullable=True),
    Column("remote_host", String(255)),
    Column("remote_addr", String(255)),
    Column("referer", TEXT),
    Column(
        "session_key", TrimmedString(255), index=True, unique=True
    ),  # unique 128 bit random number coerced to a string
    Column("is_valid", Boolean, default=False),
    Column("prev_session_id", Integer),
)  # saves a reference to the previous session so we have a way to chain them together

Tool_table = Table(
    "tool",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("guid", TrimmedString(255), index=True, unique=True),
    Column("tool_id", TrimmedString(255), index=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("newer_version_id", Integer, ForeignKey("tool.id"), nullable=True),
    Column("name", TrimmedString(255), index=True),
    Column("description", TEXT),
    Column("user_description", TEXT),
    Column("version", TrimmedString(255)),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
    Column("external_filename", TEXT),
    Column("deleted", Boolean, index=True, default=False),
)

Event_table = Table(
    "event",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("state", TrimmedString(255), index=True),
    Column("comment", TEXT),
)

ToolEventAssociation_table = Table(
    "tool_event_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("tool_id", Integer, ForeignKey("tool.id"), index=True),
    Column("event_id", Integer, ForeignKey("event.id"), index=True),
)

Category_table = Table(
    "category",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("name", TrimmedString(255), index=True, unique=True),
    Column("description", TEXT),
    Column("deleted", Boolean, index=True, default=False),
)

ToolCategoryAssociation_table = Table(
    "tool_category_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("tool_id", Integer, ForeignKey("tool.id"), index=True),
    Column("category_id", Integer, ForeignKey("category.id"), index=True),
)

Tag_table = Table(
    "tag",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("type", Integer),
    Column("parent_id", Integer, ForeignKey("tag.id")),
    Column("name", TrimmedString(255)),
    UniqueConstraint("name"),
)

ToolTagAssociation_table = Table(
    "tool_tag_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("tool_id", Integer, ForeignKey("tool.id"), index=True),
    Column("tag_id", Integer, ForeignKey("tag.id"), index=True),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
    Column("user_tname", TrimmedString(255), index=True),
    Column("value", TrimmedString(255), index=True),
    Column("user_value", TrimmedString(255), index=True),
)

ToolAnnotationAssociation_table = Table(
    "tool_annotation_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("tool_id", Integer, ForeignKey("tool.id"), index=True),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
    Column("annotation", TEXT),
)


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.create_all()
    Index(
        "ix_tool_annotation_association_annotation", ToolAnnotationAssociation_table.c.annotation, mysql_length=767
    ).create()


def downgrade(migrate_engine):
    # Operations to reverse the above upgrade go here.
    pass



"""
API operations on annotations.
"""
import logging

from galaxy.managers.display_applications import DisplayApplicationsManager
from galaxy.web import (
    expose_api,
    require_admin,
)
from . import (
    BaseGalaxyAPIController,
    depends,
)

log = logging.getLogger(__name__)


class DisplayApplicationsController(BaseGalaxyAPIController):
    manager = depends(DisplayApplicationsManager)

    @expose_api
    def index(self, trans, **kwd):
        """
        GET /api/display_applications/

        Returns the list of display applications.

        :returns:   list of available display applications
        :rtype:     list
        """
        return self.manager.index()

    @expose_api
    @require_admin
    def reload(self, trans, payload=None, **kwd):
        """
        POST /api/display_applications/reload

        Reloads the list of display applications.

        :param  ids:  list containing ids of display to be reloaded
        :type   ids:  list
        """
        payload = payload or {}
        ids = payload.get("ids", [])
        return self.manager.reload(ids)



"""
Migration script to add the includes_datatypes, has_repository_dependencies, includes_tools, includes_tool_dependencies and includes_workflows
columns to the repository_metadata table.
"""

import logging
import sys

from sqlalchemy import (
    Boolean,
    Column,
    MetaData,
    Table,
)

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Initialize.
    if migrate_engine.name == "mysql" or migrate_engine.name == "sqlite":
        default_false = "0"
    elif migrate_engine.name in ["postgres", "postgresql"]:
        default_false = "false"
    # Create and initialize tools_functionally_correct, do_not_test, time_last_tested, and tool_test_errors columns in repository_metadata table.
    RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)

    # Create tools_functionally_correct column
    c = Column("includes_datatypes", Boolean, default=False, index=True)
    try:
        c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_inc_datatypes")
        assert c is RepositoryMetadata_table.c.includes_datatypes
        migrate_engine.execute(f"UPDATE repository_metadata SET includes_datatypes={default_false}")
    except Exception:
        log.exception("Adding includes_datatypes column to the repository_metadata table failed.")

    # Create includes_datatypes column
    c = Column("has_repository_dependencies", Boolean, default=False, index=True)
    try:
        c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_has_repo_deps")
        assert c is RepositoryMetadata_table.c.has_repository_dependencies
        migrate_engine.execute(f"UPDATE repository_metadata SET has_repository_dependencies={default_false}")
    except Exception:
        log.exception("Adding has_repository_dependencies column to the repository_metadata table failed.")

    # Create includes_tools column
    c = Column("includes_tools", Boolean, default=False, index=True)
    try:
        c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_inc_tools")
        assert c is RepositoryMetadata_table.c.includes_tools
        migrate_engine.execute(f"UPDATE repository_metadata SET includes_tools={default_false}")
    except Exception:
        log.exception("Adding includes_tools column to the repository_metadata table failed.")

    # Create includes_tool_dependencies column
    c = Column("includes_tool_dependencies", Boolean, default=False, index=True)
    try:
        c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_inc_tool_deps")
        assert c is RepositoryMetadata_table.c.includes_tool_dependencies
        migrate_engine.execute(f"UPDATE repository_metadata SET includes_tool_dependencies={default_false}")
    except Exception:
        log.exception("Adding includes_tool_dependencies column to the repository_metadata table failed.")

    # Create includes_workflows column
    c = Column("includes_workflows", Boolean, default=False, index=True)
    try:
        c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_inc_workflows")
        assert c is RepositoryMetadata_table.c.includes_workflows
        migrate_engine.execute(f"UPDATE repository_metadata SET includes_workflows={default_false}")
    except Exception:
        log.exception("Adding includes_workflows column to the repository_metadata table failed.")


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop tool_test_errors, time_last_tested, do_not_test, and tools_functionally_correct columns from repository_metadata table.
    RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)

    # Drop the includes_workflows column.
    try:
        RepositoryMetadata_table.c.includes_workflows.drop()
    except Exception:
        log.exception("Dropping column includes_workflows from the repository_metadata table failed.")

    # Drop the includes_tool_dependencies column.
    try:
        RepositoryMetadata_table.c.includes_tool_dependencies.drop()
    except Exception:
        log.exception("Dropping column includes_tool_dependencies from the repository_metadata table failed.")

    # Drop the includes_tools column.
    try:
        RepositoryMetadata_table.c.includes_tools.drop()
    except Exception:
        log.exception("Dropping column includes_tools from the repository_metadata table failed.")

    # Drop the has_repository_dependencies column.
    try:
        RepositoryMetadata_table.c.has_repository_dependencies.drop()
    except Exception:
        log.exception("Dropping column has_repository_dependencies from the repository_metadata table failed.")

    # Drop the includes_datatypes column.
    try:
        RepositoryMetadata_table.c.includes_datatypes.drop()
    except Exception:
        log.exception("Dropping column includes_datatypes from the repository_metadata table failed.")



"""
Migration script to add the api_keys table.
"""

import datetime
import logging
import sys

from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    Table,
)

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import TrimmedString

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

now = datetime.datetime.utcnow

metadata = MetaData()

APIKeys_table = Table(
    "api_keys",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
    Column("key", TrimmedString(32), index=True, unique=True),
)


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    try:
        APIKeys_table.create()
    except Exception:
        log.exception("Creating api_keys table failed.")


def downgrade(migrate_engine):
    # Load existing tables
    metadata.bind = migrate_engine
    metadata.reflect()
    try:
        APIKeys_table.drop()
    except Exception:
        log.exception("Dropping api_keys table failed.")



"""
Migration script to create the repository_role_association table, insert name-spaced
repository administrative roles into the role table and associate each repository and
owner with the appropriate name-spaced role.
"""

import datetime
import logging
import sys

from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    Table,
)
from sqlalchemy.exc import NoSuchTableError

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()

NOW = datetime.datetime.utcnow
ROLE_TYPE = "system"

RepositoryRoleAssociation_table = Table(
    "repository_role_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("repository_id", Integer, ForeignKey("repository.id"), index=True),
    Column("role_id", Integer, ForeignKey("role.id"), index=True),
    Column("create_time", DateTime, default=NOW),
    Column("update_time", DateTime, default=NOW, onupdate=NOW),
)


def nextval(migrate_engine, table, col="id"):
    if migrate_engine.name in ["postgresql", "postgres"]:
        return f"nextval('{table}_{col}_seq')"
    elif migrate_engine.name in ["mysql", "sqlite"]:
        return "null"
    else:
        raise Exception(f"Unable to convert data for unknown database type: {migrate_engine.name}")


def localtimestamp(migrate_engine):
    if migrate_engine.name in ["postgresql", "postgres", "mysql"]:
        return "LOCALTIMESTAMP"
    elif migrate_engine.name == "sqlite":
        return "current_date || ' ' || current_time"
    else:
        raise Exception(f"Unable to convert data for unknown database type: {migrate_engine.name}")


def boolean_false(migrate_engine):
    if migrate_engine.name in ["postgresql", "postgres", "mysql"]:
        return False
    elif migrate_engine.name == "sqlite":
        return 0
    else:
        raise Exception(f"Unable to convert data for unknown database type: {migrate_engine.name}")


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Create the new repository_role_association table.
    try:
        RepositoryRoleAssociation_table.create()
    except Exception:
        log.exception("Creating repository_role_association table failed.")
    # Select the list of repositories and associated public user names for their owners.
    user_ids = []
    repository_ids = []
    role_names = []
    cmd = "SELECT repository.id, repository.name, repository.user_id, galaxy_user.username FROM repository, galaxy_user WHERE repository.user_id = galaxy_user.id;"
    for row in migrate_engine.execute(cmd):
        repository_id = row[0]
        name = row[1]
        user_id = row[2]
        username = row[3]
        repository_ids.append(int(repository_id))
        role_names.append(f"{str(name)}_{str(username)}_admin")
        user_ids.append(int(user_id))
    # Insert a new record into the role table for each new role.
    for tup in zip(repository_ids, user_ids, role_names):
        repository_id, user_id, role_name = tup
        cmd = "INSERT INTO role VALUES ("
        cmd += f"{nextval(migrate_engine, 'role')}, "
        cmd += f"{localtimestamp(migrate_engine)}, "
        cmd += f"{localtimestamp(migrate_engine)}, "
        cmd += f"'{role_name}', "
        cmd += "'A user or group member with this role can administer this repository.', "
        cmd += f"'{ROLE_TYPE}', "
        cmd += f"{boolean_false(migrate_engine)}"
        cmd += ");"
        migrate_engine.execute(cmd)
        # Get the id of the new role.
        cmd = f"SELECT id FROM role WHERE name = '{role_name}' and type = '{ROLE_TYPE}';"
        row = migrate_engine.execute(cmd).fetchone()
        if row:
            role_id = row[0]
        else:
            role_id = None
        if role_id:
            # Create a repository_role_association record to associate the repository with the new role.
            cmd = "INSERT INTO repository_role_association VALUES ("
            cmd += f"{nextval(migrate_engine, 'repository_role_association')}, "
            cmd += "%d, " % int(repository_id)
            cmd += "%d, " % int(role_id)
            cmd += f"{localtimestamp(migrate_engine)}, "
            cmd += f"{localtimestamp(migrate_engine)} "
            cmd += ");"
            migrate_engine.execute(cmd)
            # Create a user_role_association record to associate the repository owner with the new role.
            cmd = "INSERT INTO user_role_association VALUES ("
            cmd += f"{nextval(migrate_engine, 'user_role_association')}, "
            cmd += "%d, " % int(user_id)
            cmd += "%d, " % int(role_id)
            cmd += f"{localtimestamp(migrate_engine)}, "
            cmd += f"{localtimestamp(migrate_engine)} "
            cmd += ");"
            migrate_engine.execute(cmd)


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Determine the list of roles to delete by first selecting the list of repositories and associated
    # public user names for their owners.
    role_names = []
    cmd = "SELECT name, username FROM repository, galaxy_user WHERE repository.user_id = galaxy_user.id;"
    for row in migrate_engine.execute(cmd):
        name = row[0]
        username = row[1]
        role_names.append(f"{str(name)}_{str(username)}_admin")
    # Delete each role as well as all users associated with each role.
    for role_name in role_names:
        # Select the id of the record associated with the current role_name from the role table.
        cmd = f"SELECT id, name FROM role WHERE name = '{role_name}';"
        row = migrate_engine.execute(cmd).fetchone()
        if row:
            role_id = row[0]
        else:
            role_id = None
        if role_id:
            # Delete all user_role_association records for the current role.
            cmd = "DELETE FROM user_role_association WHERE role_id = %d;" % int(role_id)
            migrate_engine.execute(cmd)
            # Delete all repository_role_association records for the current role.
            cmd = "DELETE FROM repository_role_association WHERE role_id = %d;" % int(role_id)
            migrate_engine.execute(cmd)
            # Delete the role from the role table.
            cmd = "DELETE FROM role WHERE id = %d;" % int(role_id)
            migrate_engine.execute(cmd)
    # Drop the repository_role_association table.
    try:
        RepositoryRoleAssociation_table = Table("repository_role_association", metadata, autoload=True)
    except NoSuchTableError:
        log.debug("Failed loading table repository_role_association")
    try:
        RepositoryRoleAssociation_table.drop()
    except Exception:
        log.exception("Dropping repository_role_association table failed.")



"""
Migration script to add the email_alerts column to the repository table.
"""

import logging
import sys

from sqlalchemy import (
    Column,
    MetaData,
    Table,
)

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import JSONType

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Create and initialize imported column in job table.
    Repository_table = Table("repository", metadata, autoload=True)
    c = Column("email_alerts", JSONType, nullable=True)
    try:
        # Create
        c.create(Repository_table)
        assert c is Repository_table.c.email_alerts
    except Exception:
        log.exception("Adding email_alerts column to the repository table failed.")


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop email_alerts column from repository table.
    Repository_table = Table("repository", metadata, autoload=True)
    try:
        Repository_table.c.email_alerts.drop()
    except Exception:
        log.exception("Dropping column email_alerts from the repository table failed.")



"""
Migration script to alter the repository_metadata table by dropping the tool_test_errors column and adding columns
tool_test_results, missing_test_components.
"""

import logging
import sys

from sqlalchemy import (
    Boolean,
    Column,
    MetaData,
    Table,
)
from sqlalchemy.exc import NoSuchTableError

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import JSONType

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Initialize.
    if migrate_engine.name == "mysql" or migrate_engine.name == "sqlite":
        default_false = "0"
    elif migrate_engine.name in ["postgresql", "postgres"]:
        default_false = "false"

    try:
        RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)
    except NoSuchTableError:
        RepositoryMetadata_table = None
        log.debug("Failed loading table repository_metadata.")

    if RepositoryMetadata_table is not None:
        # Drop the tool_test_errors column from the repository_metadata table as it is poorly named.  It will be replaced with the new
        # tool_test_results column.
        try:
            col = RepositoryMetadata_table.c.tool_test_errors
            col.drop()
        except Exception:
            log.exception("Dropping column 'tool_test_errors' from repository_metadata table failed.")

        # Create the tool_test_results column to replace the ill-named tool_test_errors column just dropped above.
        c = Column("tool_test_results", JSONType, nullable=True)
        try:
            c.create(RepositoryMetadata_table)
            assert c is RepositoryMetadata_table.c.tool_test_results
        except Exception:
            log.exception("Adding tool_test_results column to the repository_metadata table failed.")

        # Create the missing_test_components column.
        c = Column("missing_test_components", Boolean, default=False, index=True)
        try:
            c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_mtc")
            assert c is RepositoryMetadata_table.c.missing_test_components
            migrate_engine.execute(f"UPDATE repository_metadata SET missing_test_components={default_false}")
        except Exception:
            log.exception("Adding missing_test_components column to the repository_metadata table failed.")


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop missing_test_components and tool_test_results from the repository_metadata table and add tool_test_errors to the repository_metadata table.
    RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)

    # Drop the missing_test_components column.
    try:
        RepositoryMetadata_table.c.missing_test_components.drop()
    except Exception:
        log.exception("Dropping column missing_test_components from the repository_metadata table failed.")

    # Drop the tool_test_results column.
    try:
        RepositoryMetadata_table.c.tool_test_results.drop()
    except Exception:
        log.exception("Dropping column tool_test_results from the repository_metadata table failed.")

    # Create the tool_test_errors column.
    c = Column("tool_test_errors", JSONType, nullable=True)
    try:
        c.create(RepositoryMetadata_table)
        assert c is RepositoryMetadata_table.c.tool_test_errors
    except Exception:
        log.exception("Adding tool_test_errors column to the repository_metadata table failed.")



"""
Migration script to add the malicious column to the repository_metadata table.
"""

import logging
import sys

from sqlalchemy import (
    Boolean,
    Column,
    MetaData,
    Table,
)

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Create and initialize imported column in job table.
    Repository_metadata_table = Table("repository_metadata", metadata, autoload=True)
    c = Column("malicious", Boolean, default=False, index=True)
    try:
        # Create
        c.create(Repository_metadata_table, index_name="ix_repository_metadata_malicious")
        assert c is Repository_metadata_table.c.malicious
        # Initialize.
        if migrate_engine.name == "mysql" or migrate_engine.name == "sqlite":
            default_false = "0"
        elif migrate_engine.name in ["postgresql", "postgres"]:
            default_false = "false"
        migrate_engine.execute(f"UPDATE repository_metadata SET malicious={default_false}")
    except Exception:
        log.exception("Adding malicious column to the repository_metadata table failed.")


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop malicious column from repository_metadata table.
    Repository_metadata_table = Table("repository_metadata", metadata, autoload=True)
    try:
        Repository_metadata_table.c.malicious.drop()
    except Exception:
        log.exception("Dropping column malicious from the repository_metadata table failed.")



"""
Adds the repository, repository_rating_association and repository_category_association tables.
"""

import datetime
import logging
import sys

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    Table,
    TEXT,
)

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import TrimmedString

now = datetime.datetime.utcnow
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()

Repository_table = Table(
    "repository",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("name", TrimmedString(255), index=True),
    Column("description", TEXT),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
    Column("private", Boolean, default=False),
    Column("deleted", Boolean, index=True, default=False),
)

RepositoryRatingAssociation_table = Table(
    "repository_rating_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("repository_id", Integer, ForeignKey("repository.id"), index=True),
    Column("user_id", Integer, ForeignKey("galaxy_user.id"), index=True),
    Column("rating", Integer, index=True),
    Column("comment", TEXT),
)

RepositoryCategoryAssociation_table = Table(
    "repository_category_association",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("repository_id", Integer, ForeignKey("repository.id"), index=True),
    Column("category_id", Integer, ForeignKey("category.id"), index=True),
)


def upgrade(migrate_engine):
    print(__doc__)
    # Load existing tables
    metadata.bind = migrate_engine
    metadata.reflect()
    try:
        Repository_table.create()
    except Exception:
        log.exception("Creating repository table failed.")
    try:
        RepositoryRatingAssociation_table.create()
    except Exception:
        log.exception("Creating repository_rating_association table failed.")
    try:
        RepositoryCategoryAssociation_table.create()
    except Exception:
        log.exception("Creating repository_category_association table failed.")


def downgrade(migrate_engine):
    # Load existing tables
    metadata.bind = migrate_engine
    metadata.reflect()
    try:
        Repository_table.drop()
    except Exception:
        log.exception("Dropping repository table failed.")
    try:
        RepositoryRatingAssociation_table.drop()
    except Exception:
        log.exception("Dropping repository_rating_association table failed.")
    try:
        RepositoryCategoryAssociation_table.drop()
    except Exception:
        log.exception("Dropping repository_category_association table failed.")



from typing import TYPE_CHECKING

from galaxy.structured_app import BasicSharedApp

if TYPE_CHECKING:
    from tool_shed.repository_types.registry import Registry as RepositoryTypesRegistry
    from tool_shed.webapp.model import mapping


class ToolShedApp(BasicSharedApp):
    repository_types_registry: "RepositoryTypesRegistry"
    model: "mapping.ToolShedModelMapping"



"""
Migration script to add the remote_repository_url and homepage_url
columns to the repository table.
"""

import logging
import sys

from sqlalchemy import (
    Column,
    MetaData,
    Table,
)

from galaxy.model.custom_types import TrimmedString

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

metadata = MetaData()


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    Repository_table = Table("repository", metadata, autoload=True)
    c_remote = Column("remote_repository_url", TrimmedString(255))
    c_homepage = Column("homepage_url", TrimmedString(255))
    try:
        # Create
        c_remote.create(Repository_table)
        c_homepage.create(Repository_table)
        assert c_remote is Repository_table.c.remote_repository_url
        assert c_homepage is Repository_table.c.homepage_url
    except Exception:
        log.exception("Adding remote_repository_url and homepage_url columns to the repository table failed.")


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()
    # Drop type column from repository table.
    Repository_table = Table("repository", metadata, autoload=True)
    try:
        Repository_table.c.remote_repository_url.drop()
        Repository_table.c.homepage_url.drop()
    except Exception:
        log.exception("Dropping columns remote_repository_url and homepage_url from the repository table failed.")



"""
Migration script to add the skip_tool_test table and add the test_install_error column to the repository_metadata table.
"""

import datetime
import logging
import sys

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    MetaData,
    Table,
    TEXT,
)
from sqlalchemy.exc import NoSuchTableError

# Need our custom types, but don't import anything else from model
from galaxy.model.custom_types import TrimmedString

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
format = "%(name)s %(levelname)s %(asctime)s %(message)s"
formatter = logging.Formatter(format)
handler.setFormatter(formatter)
log.addHandler(handler)

now = datetime.datetime.utcnow

metadata = MetaData()

SkipToolTest_table = Table(
    "skip_tool_test",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("create_time", DateTime, default=now),
    Column("update_time", DateTime, default=now, onupdate=now),
    Column("repository_metadata_id", Integer, ForeignKey("repository_metadata.id"), index=True),
    Column("initial_changeset_revision", TrimmedString(255), index=True),
    Column("comment", TEXT),
)


def upgrade(migrate_engine):
    print(__doc__)
    metadata.bind = migrate_engine
    metadata.reflect()
    # Initialize.
    if migrate_engine.name == "mysql" or migrate_engine.name == "sqlite":
        default_false = "0"
    elif migrate_engine.name in ["postgresql", "postgres"]:
        default_false = "false"

    try:
        RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)
    except NoSuchTableError:
        RepositoryMetadata_table = None
        log.debug("Failed loading table repository_metadata.")

    if RepositoryMetadata_table is not None:
        # Create the test_install_error column.
        c = Column("test_install_error", Boolean, default=False, index=True)
        try:
            c.create(RepositoryMetadata_table, index_name="ix_repository_metadata_ttie")
            assert c is RepositoryMetadata_table.c.test_install_error
            migrate_engine.execute(f"UPDATE repository_metadata SET test_install_error={default_false}")
        except Exception:
            log.exception("Adding test_install_error column to the repository_metadata table failed.")

    # Create skip_tool_test table.
    try:
        SkipToolTest_table.create()
    except Exception:
        log.exception("Creating the skip_tool_test table failed.")


def downgrade(migrate_engine):
    metadata.bind = migrate_engine
    metadata.reflect()

    # Drop the skip_tool_test table.
    try:
        SkipToolTest_table.drop()
    except Exception:
        log.exception("Dropping the skip_tool_test table failed.")

    # Drop test_install_error column from the repository_metadata table.
    RepositoryMetadata_table = Table("repository_metadata", metadata, autoload=True)
    try:
        RepositoryMetadata_table.c.test_install_error.drop()
    except Exception:
        log.exception("Dropping column test_install_error from the repository_metadata table failed.")
